local ffi = require 'ffi'
local lj_include = require 'lj_include'

lj_include('stdint')
lj_include('sys_socket')

ffi.cdef[[
typedef uint32_t in_addr_t;
struct in_addr {
    in_addr_t s_addr;
};
]]

lj_include('bits_in')

ffi.cdef[[
enum {
    IPPROTO_IP = 0,
    IPPROTO_ICMP = 1,
    IPPROTO_IGMP = 2,
    IPPROTO_IPIP = 4,
    IPPROTO_TCP = 6,
    IPPROTO_EGP = 8,
    IPPROTO_PUP = 12,
    IPPROTO_UDP = 17,
    IPPROTO_IDP = 22,
    IPPROTO_TP = 29,
    IPPROTO_DCCP = 33,
    IPPROTO_IPV6 = 41,
    IPPROTO_RSVP = 46,
    IPPROTO_GRE = 47,
    IPPROTO_ESP = 50,
    IPPROTO_AH = 51,
    IPPROTO_MTP = 92,
    IPPROTO_BEETPH = 94,
    IPPROTO_ENCAP = 98,
    IPPROTO_PIM = 103,
    IPPROTO_COMP = 108,
    IPPROTO_SCTP = 132,
    IPPROTO_UDPLITE = 136,
    IPPROTO_RAW = 255,
    IPPROTO_MAX
};
enum {
    IPPROTO_HOPOPTS = 0,
    IPPROTO_ROUTING = 43,
    IPPROTO_FRAGMENT = 44,
    IPPROTO_ICMPV6 = 58,
    IPPROTO_NONE = 59,
    IPPROTO_DSTOPTS = 60,
    IPPROTO_MH = 135
};
typedef uint16_t in_port_t;
enum {
    IPPORT_ECHO = 7,
    IPPORT_DISCARD = 9,
    IPPORT_SYSTAT = 11,
    IPPORT_DAYTIME = 13,
    IPPORT_NETSTAT = 15,
    IPPORT_FTP = 21,
    IPPORT_TELNET = 23,
    IPPORT_SMTP = 25,
    IPPORT_TIMESERVER = 37,
    IPPORT_NAMESERVER = 42,
    IPPORT_WHOIS = 43,
    IPPORT_MTP = 57,
    IPPORT_TFTP = 69,
    IPPORT_RJE = 77,
    IPPORT_FINGER = 79,
    IPPORT_TTYLINK = 87,
    IPPORT_SUPDUP = 95,
    IPPORT_EXECSERVER = 512,
    IPPORT_LOGINSERVER = 513,
    IPPORT_CMDSERVER = 514,
    IPPORT_EFSSERVER = 520,
    IPPORT_BIFFUDP = 512,
    IPPORT_WHOSERVER = 513,
    IPPORT_ROUTESERVER = 520,
    IPPORT_RESERVED = 1024,
    IPPORT_USERRESERVED = 5000
};
struct in6_addr {
    union {
	uint8_t __u6_addr8[16];
	uint16_t __u6_addr16[8];
	uint32_t __u6_addr32[4];
    } __in6_u;
};
const struct in6_addr in6addr_any;
const struct in6_addr in6addr_loopback;
struct sockaddr_in {
    sa_family_t sin_family;
    in_port_t sin_port;
    struct in_addr sin_addr;
    unsigned char sin_zero[sizeof(struct sockaddr) -
	(sizeof(unsigned short int)) -
	sizeof(in_port_t) - sizeof(struct in_addr)];
};
struct sockaddr_in6 {
    sa_family_t sin6_family;
    in_port_t sin6_port;
    uint32_t sin6_flowinfo;
    struct in6_addr sin6_addr;
    uint32_t sin6_scope_id;
};
struct ip_mreq {
    struct in_addr imr_multiaddr;
    struct in_addr imr_interface;
};
struct ip_mreq_source {
    struct in_addr imr_multiaddr;
    struct in_addr imr_interface;
    struct in_addr imr_sourceaddr;
};
struct ipv6_mreq {
    struct in6_addr ipv6mr_multiaddr;
    unsigned int ipv6mr_interface;
};
struct group_req {
    uint32_t gr_interface;
    struct sockaddr_storage gr_group;
};
struct group_source_req {
    uint32_t gsr_interface;
    struct sockaddr_storage gsr_group;
    struct sockaddr_storage gsr_source;
};
struct ip_msfilter {
    struct in_addr imsf_multiaddr;
    struct in_addr imsf_interface;
    uint32_t imsf_fmode;
    uint32_t imsf_numsrc;
    struct in_addr imsf_slist[1];
};
struct group_filter {
    uint32_t gf_interface;
    struct sockaddr_storage gf_group;
    uint32_t gf_fmode;
    uint32_t gf_numsrc;
    struct sockaddr_storage gf_slist[1];
};
uint32_t ntohl(uint32_t __netlong);
uint16_t ntohs(uint16_t __netshort);
uint32_t htonl(uint32_t __hostlong);
uint16_t htons(uint16_t __hostshort);
]]

lj_include('bits_byteswap')

ffi.cdef[[
int bindresvport(int __sockfd, struct sockaddr_in *__sock_in);
int bindresvport6(int __sockfd, struct sockaddr_in6 *__sock_in);
struct cmsghdr;
struct in6_pktinfo {
    struct in6_addr ipi6_addr;
    unsigned int ipi6_ifindex;
};
struct ip6_mtuinfo {
    struct sockaddr_in6 ip6m_addr;
    uint32_t ip6m_mtu;
};
int inet6_option_space(int __nbytes);
int inet6_option_init(void *__bp, struct cmsghdr **__cmsgp, int __type);
int inet6_option_append(struct cmsghdr *__cmsg,
    const uint8_t * __typep, int __multx, int __plusy);
uint8_t *inet6_option_alloc(struct cmsghdr *__cmsg, int __datalen,
    int __multx, int __plusy);
int inet6_option_next(const struct cmsghdr *__cmsg, uint8_t ** __tptrp);
int inet6_option_find(const struct cmsghdr *__cmsg,
    uint8_t ** __tptrp, int __type);
int inet6_opt_init(void *__extbuf, socklen_t __extlen);
int inet6_opt_append(void *__extbuf, socklen_t __extlen, int __offset,
    uint8_t __type, socklen_t __len, uint8_t __align, void **__databufp);
int inet6_opt_finish(void *__extbuf, socklen_t __extlen, int __offset);
int inet6_opt_set_val(void *__databuf, int __offset, void *__val,
    socklen_t __vallen);
int inet6_opt_next(void *__extbuf, socklen_t __extlen, int __offset,
    uint8_t * __typep, socklen_t * __lenp, void **__databufp);
int inet6_opt_find(void *__extbuf, socklen_t __extlen, int __offset,
    uint8_t __type, socklen_t * __lenp, void **__databufp);
int inet6_opt_get_val(void *__databuf, int __offset, void *__val,
    socklen_t __vallen);
socklen_t inet6_rth_space(int __type, int __segments);
void *inet6_rth_init(void *__bp, socklen_t __bp_len, int __type,
    int __segments);
int inet6_rth_add(void *__bp, const struct in6_addr *__addr);
int inet6_rth_reverse(const void *__in, void *__out);
int inet6_rth_segments(const void *__bp);
struct in6_addr *inet6_rth_getaddr(const void *__bp, int __index);
int getipv4sourcefilter(int __s, struct in_addr __interface_addr,
    struct in_addr __group, uint32_t * __fmode,
    uint32_t * __numsrc, struct in_addr *__slist);
int setipv4sourcefilter(int __s, struct in_addr __interface_addr,
    struct in_addr __group, uint32_t __fmode,
    uint32_t __numsrc, const struct in_addr *__slist);
int getsourcefilter(int __s, uint32_t __interface_addr,
    const struct sockaddr *__group,
    socklen_t __grouplen, uint32_t * __fmode,
    uint32_t * __numsrc, struct sockaddr_storage *__slist);
int setsourcefilter(int __s, uint32_t __interface_addr,
    const struct sockaddr *__group,
    socklen_t __grouplen, uint32_t __fmode,
    uint32_t __numsrc, const struct sockaddr_storage *__slist);
]]

local defs = {
    IN_CLASSA_NET = '0xff000000',
    IN_CLASSA_NSHIFT = '24',
    IN_CLASSA_MAX = '128',
    IN_CLASSB_NET = '0xffff0000',
    IN_CLASSB_NSHIFT = '16',
    IN_CLASSB_MAX = '65536',
    IN_CLASSC_NET = '0xffffff00',
    IN_CLASSC_NSHIFT = '8',
    IN_LOOPBACKNET = '127',
    INET_ADDRSTRLEN = '16',
    INET6_ADDRSTRLEN = '46',
}
local format, cdef, C = string.format, ffi.cdef, ffi.C
for sym, val in pairs(defs) do
    if not pcall(function() return C[sym] end) then
        cdef(format("enum {%s = %s};", sym, val))
    end
end

