--[[
    Holds majority of the sxr import rules, per document SomatXR HBM SDB Sensor Import (Rev 10), 25-Oct-15 (DFD)
    supported_bridges are the supported bridge type enums
    scale_types are the different types of sensor scaling we support. Each has it's own set of rules.
    same for ss. Each supported sensor has a number of scaling types associated with it, hence the previous data struct.
--]]

local ss

--================== Supported Bridge Bridge Enumerations =====
local supported_bridges =
    { 362 ,363 ,364 ,365 ,366 ,367 ,368 ,369 ,371 ,372 ,350 ,351 ,353 ,354, 355}

local bridge_map = {
    [350] = {mode = "bridge", type = "full"},
    [351] = {mode = "bridge", type = "half"},
    [353] = {mode = "bridge", type = "full"},
    [354] = {mode = "bridge", type = "half"},
    [355] = {mode = "bridge", type = "quarter"},
    [362] = {mode = "bridge", type = "full"},
    [363] = {mode = "bridge", type = "full"},
    [364] = {mode = "bridge", type = "full"},
    [365] = {mode = "bridge", type = "half"},
    [366] = {mode = "bridge", type = "half"},
    [367] = {mode = "bridge", type = "half"},
    [368] = {mode = "bridge", type = "quarter", wires = 4},
    [369] = {mode = "bridge", type = "quarter", wires = 4},
    [371] = {mode = "bridge", type = "quarter", wires = 3},
    [372] = {mode = "bridge", type = "quarter", wires = 3}
}

local bridge_resist_map = {
    [362] = 120,
    [363] = 350,
    [364] = 700,
    [365] = 120,
    [366] = 350,
    [367] = 700,
    [368] = 120,
    [369] = 350,
    [371] = 120,
    [372] = 350
}

local mx1601 = "source/hbm_mx1601b"
local mx1615 = "source/hbm_mx1615b"
local mx1609 = "source/hbm_mx1609b"
local mx840 = "source/hbm_mx840b"
local mx411 = "source/hbm_mx411b"
local hls   = "source/hbm_edaqxr/hls"
local itb   = "source/hbm_edaqxr/itb"
local ntb   = "source/hbm_edaqxr/ntb"
local brg   = "source/hbm_edaqxr/brg"
local dio   = "source/hbm_edaqxr/dio/pulse"

--loops and calls all overlaps -> categories -> fields
--for _category_specific setup
--there's some extra stuff here to check for successive failures
--in the same field- since we can only fail a sensor once
--it's failed in the same field for each category (if multiple)
local function setup_category_specifics( sensor ,sxr_row ,args)
--    print("sensor: " .. tostring(sensor))
    sxr_row._category_specific = {}
    local failures = {}
    print("supported_categories:\t" .. #sxr_row._supported_categories) 
    for category ,fields in pairs( ss[ sensor].overlaps) do
        local supported = false
        local needs_sensor_input = false
        for k,v in pairs(sxr_row._supported_categories) do
            if category == v then
                supported = true
            end
        end
        sxr_row._category_specific[ category] = {}
        sxr_row._category_specific[ category].sensor_input = copy_table(sxr_row.sensor_input)
        if supported then
            for field ,fn in pairs( fields) do
                if field ==  "sensor_input" then needs_sensor_input = true end
                local success = fn( sxr_row ,args)
                if (not success) then
                    print("failed:\t" .. category)
                    failures[ category] = true
                    break
                end
            end
        else
            failures[category] = true
        end
        if not needs_sensor_input then 
            sxr_row._category_specific[ category].sensor_input = nil
        end
    end
    for category, fields in pairs (ss[sensor].overlaps) do
        for field, fn in pairs (fields) do
            sxr_row[ field] = nil 
            --moved here so common parameters can be done in the body and copied to overlaps
        end
    end
    --print("Category Count: " ,cc)

    for category,_ in pairs( failures) do
        sxr_row._category_specific[ category] = nil
        for k,v in pairs(sxr_row._supported_categories) do
            if v == category then
                table.remove(sxr_row._supported_categories, k)
                break
            end
        end
    end
    print("supported_categories:\t" .. #sxr_row._supported_categories) 
    return #sxr_row._supported_categories == 0
end


--================== Supported Scale Types ======================
local scale_types = {
    StrainGage = {    enums = {0}
        ,sxr_name="strain_gage"
        ,required = function( cal ,l)
            local failed = false
            cal.gage_factor = l.GageFactor
            cal.bridge_factor = l.BridgeFactor
            if cal.gage_factor < 0.1
                or cal.gage_factor > 10
                or cal.bridge_factor == 0 then
                failed = true
                print( "Failed strain gage check.")
            end
            return failed
        end
    }

    ,ZeroSpan = {     enums = {0,1}
        ,sxr_name="defined_zero_span"
        ,required = function( cal ,l)
            local failed = false
            cal.electrical_zero = l.P1X
            cal.electrical_span = l.P2X - l.P1X
            cal.physical_span = l.P2Y
            if cal.electrical_span == 0 or l.P2Y == 0 then
                failed = true
                print( "Failed ZeroSpan check." )
            end
            --cal.max_electrical =
          --      math.abs(cal.electrical_span) +
            --      math.abs(cal.electrical_zero)
            return failed
        end
    }

    ,Table2 = {   enums = {2}
        ,sxr_name = "defined_two_point"
        ,required = function( cal , l )
            local failed = false
            --print( "Scale data: " ,l.ScaleData1 )
            local i,e1,p1,e2,p2 = string.match( l.ScaleData1
                ,'([-.%d]+)~([-.%d]+)~([-.%d]+)~([-.%d]+)~([-.%d]+)' )
            --print(tostring(i).."\t"..tostring(e1).."\t"..tostring(p1).."\t"..tostring(e2).."\t"..tostring(p2))
            if i ~= '4' then
                print( 'Failed Table2 2Point check, ',i )
                return true
            end
            cal.electrical = {}
            cal.physical = {}
            cal.electrical[1] = tonumber( e1 )
            cal.electrical[2] = tonumber( e2 )
            cal.physical[1] = tonumber( p1 )
            cal.physical[2] = tonumber( p2 )
            --print("CHECK: " , e1 , e2 , p1 , p2 )
            if e1 == e2 or p1 == p2 then
                failed = true
                print( "Failed Table2 check." )
            end
            return failed
        end
    }

    ,TwoPoint = {     enums = {32}
        ,sxr_name="defined_two_point"
        ,required = function( cal , l )
            local failed = false
            cal.electrical = {}
            cal.electrical[1] = l.P1X
            cal.electrical[2] = l.P2X
            cal.physical = {}
            cal.physical[1] = l.P1Y
            cal.physical[2] = l.P2Y
            if l.P1X == l.P2X or l.P1Y == l.P2Y then
                failed = true
                print( "Failing TwoPoint check.")
            end
            return failed
        end
    }

    ,Gain =       {   enums = {64}
        ,sxr_name = "defined_slope_intercept"
        ,required = function( cal , l )
            --print( l.ScaleData5)
            cal.intercept ,cal.slope = l.ScaleData5:match(  '([%d%.]+)~([%d%.]+)')
            cal.intercept = tonumber( cal.intercept)
            cal.slope = tonumber( cal.slope)
            return false
        end
    }

    ,Internal =   {   enums = {0,256,512}
        ,sxr_name = "internal"
        ,required = function( cal , l ) --internal has no requirements, per spec's given by Dennis
            return false
        end
    }

}

--==================== Supported Sensor Types =======================================

--Supported Sensors
ss = {}

ss.StrainGage = {
    name = "StrainGage"
    ,enum = 1
    ,elec_units = "mV/V"
    ,s_types = {
        scale_types.ZeroSpan
        ,scale_types.StrainGage
        ,scale_types.TwoPoint
        ,scale_types.Table2
        ,scale_types.Gain
    }
    ,overlaps = {
        [mx1615] = {
            sensor_input = function (s , args)
                --s._category_specific[mx1615].sensor_input = s.sensor_input 
                return true
            end
            ,excitation_range = function( s ,args)
                local l = args.l
                local er
                local sdb_ex
                local ex
                local max
                local ref = {}

                if l.Resistance == 0 then l.Resistance = 350
                end
                if l.ScaleType == 0 then
                    if l.GageFactor == 0 then -- ZeroSpan
                        max = math.abs(l.P2X-l.P1X) + math.abs(l.P1X)
                        s.calibration[3] = nil
                    else -- StrainGage
                        if l.Range > 100 then max = 4
                        else max = l.Range end
                    end
                elseif  l.ScaleType == 64 then --Gain
                    max = l.Range
                    s.calibration[3] = nil
                elseif l.ScaleType == 1 then --ZeroSpan
                    max = math.abs(l.P2X-l.P1X) + math.abs(l.P1X)
                    s.calibration[3] = nil
                elseif l.ScaleType == 2 or l.ScaleType == 32 then -- Table2 or  TwoPoint
                    max = l.P2X
                    s.calibration[3] = nil
                else
                    return false
                end

                if l.Excitation >= 5 then sdb_ex = 5.0
                elseif l.Excitation >= 2.5 then sdb_ex = 2.5
                elseif l.Excitation >= 1.0 then sdb_ex = 1.0
                elseif l.Excitation >= 0.5 then sdb_ex = 0.5
                else
                    return false
                end

                if max > 40 then
                    return false
                elseif max > 20 then
                    if s.sensor_input["type"] =="quarter" then 
                        ex = 0.5
                    else
                        return false
                    end
                elseif max > 8 then
                    if sdb_ex < 1.0 then ex = sdb_ex
                    else ex = 1.0
                    end
                elseif max > 4 then
                    if sdb_ex < 2.5 then ex = sdb_ex
                    else ex = 2.5
                    end
                else
                    if sdb_ex < 5.0 then ex = sdb_ex
                    else ex = 5.0
                    end
                end

                if s.bridge_resistance < 110 and ex > 1.0 then ex = 1.0
                elseif s.bridge_resistance < 300 and ex > 2.5 then ex = 2.5
                end

                if s.sensor_input["type"] ~= "quarter" and ex == 0.5 then 
                    return false
                end

                if ex == 5.0 then
                    er = "5.0v"
                    ref.max = 4
                    ref.min = -4
                elseif ex == 2.5 then
                    er = "2.5v"
                    ref.max = 8
                    ref.min = -8
                elseif ex == 1.0 then
                    er = "1.0v"
                    ref.max = 20
                    ref.min = -20
                else
                    er = ex .. "v"
                    ref.max = 40
                    ref.min = -40
                end
                s._category_specific[ mx1615].excitation_range = er
                s._category_specific[ mx1615].sensor_input.max = ref.max
                s._category_specific[ mx1615].sensor_input.min = ref.min
                s._category_specific[ mx1615].sensor_input.wires = s.sensor_input.wires
                s._category_specific[ mx1615].sensor_input.regulation= s.sensor_input.regulation
                return true
            end
            ,excitation_mode = function( s ,args)
                local l = args.l
                local em
                if l.ExcitationFrequency == -1 then
                    em = 'dc'
                else
                    if s.sensor_input.type == "quarter" and
                        s.sensor_input.wires == "3" and
                        s.sensor_input.regulation == true then
                        em = 'dc'
                    else
                        em = 'ac_1200_hz_square'
                    end
                end
                s._category_specific[ mx1615].excitation_mode = em
                return true
            end
        }
        ,[mx840] = {
            sensor_input = function (s, args)
                --s._category_specific[mx840].sensor_input = s.sensor_input
                return true
            end
            ,excitation_range = function( s ,args)
                local l = args.l
                local er
                local sdb_ex
                local ex
                local gain_low, gain_low_low = false, false
                local max
                local t = l.SensorType
                local ref = {}

                if l.ScaleType == 0 then
                    if l.GageFactor == 0 then -- ZeroSpan
                        max = math.abs(l.P2X-l.P1X) + math.abs(l.P1X)
                        s.calibration[3] = nil
                    else -- StrainGage
                        if l.Range > 100 then max = 4
                        else max = l.Range end
                    end
                elseif  l.ScaleType == 64 then --Gain
                    max = l.Range
                    s.calibration[3] = nil
                elseif l.ScaleType == 1 then --ZeroSpan
                    max = math.abs(l.P2X - l.P1X) + math.abs(l.P1X)
                    s.calibration[3] = nil
                elseif l.ScaleType == 2 or l.ScaleType == 32 then -- Table2 or  TwoPoint
                    max = l.P2X
                    s.calibration[3] = nil
                else
                    return false
                end

                if l.ExcitationFrequency == -1 and s.sensor_input["type"] == "full" then--full/dc
                    if l.Excitation >= 2.5 then
                        sdb_ex = 2.5
                    elseif l.Excitation >= 1.0 then
                        sdb_ex = 1.0
                    else
                        return false
                    end

                    if max > 1000 then
                        return false
                    elseif max > 120 then
                        gain_low_low = true
                        if sdb_ex < 2.5 then
                            return false
                        end
                        ex = 2.5
                    elseif max > 5 then
                        gain_low = true
                        if sdb_ex < 2.5 then
                            return false
                        end
                        ex = 2.5
                    else
                        if sdb_ex < 2.5 then
                            ex = sdb_ex
                        else
                            ex = 2.5
                        end
                    end

                    if gain_low or gain_low_low then
                        if s.bridge_resistance < 300 then
                            --print("gain_low and resistance < 300")
                            return false
                        end
                    else
                        if s.bridge_resistance < 300 then
                            if ex > 1.0 then
                                ex = 1.0
                            end
                        end
                    end

                    if ex == 2.5 then
                        if gain_low_low then
                            ref.max = 1000
                            ref.min = -1000
                        elseif gain_low then
                            ref.max = 120
                            ref.min = -120
                        else
                            ref.max = 5
                            ref.min = -5
                        end
                    else
                        ref.max = 12.5
                        ref.min = -12.5
                    end

                    if ex == 2.5 then er = "2.5v"
                    else er = "1.0v"
                    end

                    s._category_specific[ mx840].excitation_range = er
                    s._category_specific[ mx840].sensor_input.max = ref.max
                    s._category_specific[ mx840].sensor_input.min = ref.min
                    s._category_specific[ mx840].sensor_input.wires = s.sensor_input.wires
                    s._category_specific[ mx840].sensor_input.regulation= s.sensor_input.regulation
                    return true
                elseif l.ExcitationFrequency == -1 and s.sensor_input["type"] == "half" then
                    if l.Excitation >= 2.5 then
                        sdb_ex = 2.5
                    elseif l.Excitation >= 1.0 then
                        sdb_ex = 1.0
                    else
                        return false
                    end

                    if max > 12.5 then
                        return false
                    elseif max > 5 then
                        ex = 1.0
                    else
                        if sdb_ex < 2.5 then
                            ex = sdb_ex
                        else
                            ex = 2.5
                        end
                    end

                    if s.bridge_resistance < 300 then
                        if ex > 1.0 then
                            ex = 1.0
                        end
                    end

                    if ex == 2.5 then
                        ref.max = 5
                        ref.min = -5
                    else
                        ref.max = 12.5
                        ref.min = -12.5
                    end

                    if ex == 2.5 then er = "2.5v"
                    else er = "1.0v"
                    end

                    s._category_specific[ mx840].excitation_range = er
                    s._category_specific[ mx840].sensor_input.max = ref.max
                    s._category_specific[ mx840].sensor_input.min = ref.min
                    s._category_specific[ mx840].sensor_input.wires = s.sensor_input.wires
                    s._category_specific[ mx840].sensor_input.regulation= s.sensor_input.regulation
                    return true
                elseif s.sensor_input["type"] == "half" or s.sensor_input["type"] == "full" then --full or half/ac
                    if l.Excitation >= 2.5 then
                        sdb_ex = 2.5
                    elseif l.Excitation >= 1.0 then
                        sdb_ex = 1.0
                    else
                        return false
                    end

                    if max > 300 then
                        return false
                    elseif max > 120 then
                        gain_low = true
                        ex = 1.0
                    elseif max > 12.5 then
                        gain_low = true
                        if sdb_ex < 2.5 then
                            ex = sdb_ex
                        else
                            ex = 2.5
                        end
                    elseif max > 5 then
                        ex = 2.5
                    else
                        ex = sdb_ex
                    end
                    --print("Resistance:\t" .. tostring(l.Resistance) .. "\tex\t" .. tostring(ex) .. "1")

                    if s.bridge_resistance < 110 then
                        if ex > 1.0 then ex = 1.0 end
                    elseif s.bridge_resistance < 300 then
                        if ex > 2.5 then ex = 2.5 end
                    end

                    if ex == 2.5 then
                        if gain_low then
                            ref.max = 120
                            ref.min = -120
                        else
                            ref.max = 5
                            ref.min = -5
                        end
                    else
                        if gain_low then
                            ref.max = 300
                            ref.min = -300
                        else
                            ref.max = 12.5
                            ref.min = -12.5
                        end
                    end

                    if ex == 2.5 then er = "2.5v"
                    else er = "1.0v"
                    end

                    s._category_specific[ mx840].excitation_range = er
                    s._category_specific[ mx840].sensor_input.max = ref.max
                    s._category_specific[ mx840].sensor_input.min = ref.min
                    s._category_specific[ mx840].sensor_input.wires = s.sensor_input.wires
                    s._category_specific[ mx840].sensor_input.regulation= s.sensor_input.regulation
                    return true
                elseif s.sensor_input["type"] == "quarter" then 
                    if l.Excitation >= 2.5 then
                        sdb_ex = 2.5
                    elseif l.Excitation >= 1.0 then
                        sdb_ex = 1.0
                    else
                        return false
                    end

                    if l.ExcitationFrequency == -1 and max > 120 then
                        --print("max > 120 with dc excitation")
                        return false
                    end
                    if max > 300 then
                        return false
                    elseif max > 120 then
                        gain_low = true
                        ex = 1.0
                    elseif max > 12.5 then
                        gain_low = true
                        if sdb_ex < 2.5 then
                            ex = sdb_ex
                        else
                            ex = 2.5
                        end
                    elseif max > 5 then
                        ex = 2.5
                    else
                        ex = sdb_ex
                    end

                    if ex == 2.5 then
                        if gain_low then
                            ref.max = 120
                            ref.min = -120
                        else
                            ref.max = 5
                            ref.min = -5
                        end
                    else
                        if gain_low then
                            ref.max = 300
                            ref.min = -300
                        else
                            ref.max = 12.5
                            ref.min = -12.5
                        end
                    end

                    if ex == 2.5 then er = "2.5v"
                    else er = "1.0v"
                    end

                    s._category_specific[ mx840].excitation_range = er
                    s._category_specific[ mx840].sensor_input.max = ref.max
                    s._category_specific[ mx840].sensor_input.min = ref.min
                    s._category_specific[ mx840].sensor_input.wires = s.sensor_input.wires
                    s._category_specific[ mx840].sensor_input.regulation= s.sensor_input.regulation
                    return true
                end
            end
            ,excitation_mode = function( s ,args)
                local l = args.l
                local em
                if l.ExcitationFrequency == -1 then
                    em = 'dc'
                else
                    em = 'ac_4800_hz_sine'
                end
                s._category_specific[ mx840].excitation_mode = em
                return true
            end
        }
        ,[mx411] = {
            sensor_input = function (s, args)
                --s._category_specific[mx411].sensor_input = s.sensor_input
                return true
            end
            ,excitation_range = function( s ,args)
                local l = args.l
                local er
                local sdb_ex
                local ex
                local gain_low = false
                local max
                local ref = {}

                if l.ScaleType == 0 then
                    if l.GageFactor == 0 then -- ZeroSpan
                        max = math.abs(l.P2X-l.P1X) + math.abs(l.P1X)
                        s.calibration[3] = nil
                    else -- StrainGage
                        if l.Range > 100 then max = 4
                        else max = l.Range end
                    end
                elseif  l.ScaleType == 64 then --Gain
                    max = l.Range
                    s.calibration[3] = nil
                elseif l.ScaleType == 1 then --ZeroSpan
                    max = math.abs(l.P2X - l.P1X) + math.abs(l.P1X)
                    s.calibration[3] = nil
                elseif l.ScaleType == 2 or l.ScaleType == 32 then -- Table2 or  TwoPoint
                    s.calibration[3] = nil
                    max = l.P2X
                else
                    return false
                end

                if l.ExcitationFrequency == -1 then
                    if l.Excitation >= 7.5 then
                        if t == 368 or t == 369 or t == 371 or t == 372 or t == 352 then --quarter bridge
                            sdb_ex = 5.0
                        else
                            sdb_ex = 7.5
                        end
                    elseif l.Excitation >= 5 then
                        sdb_ex = 5.0
                    elseif l.Excitation >= 2.5 then
                        sdb_ex = 2.5
                    elseif l.Excitation >= 1.0 then
                        sdb_ex = 1.0
                    else
                        return false
                    end

                    if max > 100 then
                        return false
                    elseif max > 50 then
                        gain_low = true
                        if sdb_ex < 2.5 then
                            return false
                        end
                        ex = 2.5
                    elseif max > 20 then
                        gain_low = true
                        if sdb_ex < 5.0 then
                            ex = sdb_ex
                            if sdb_ex < 2.5 then
                                return false
                            end
                        else
                            ex = 5.0
                        end
                    elseif max > 10 then
                        ex = 1.0
                    elseif max > 4 then
                        if sdb_ex < 2.5 then
                            ex = sdb_ex
                        else
                            ex = 2.5
                        end
                    else
                        if sdb_ex < 7.5 then
                            ex = sdb_ex
                        else
                            ex = 7.5
                        end
                    end

                    if s.bridge_resistance < 110 then
                        if ex > 1.0 then
                            ex = 1.0
                        end
                    elseif s.bridge_resistance < 300 then
                        if ex > 5.0 then
                            ex = 5.0
                        end
                    end

                    if ex == 7.5 then
                        ref.max = 4
                        ref.min = -4
                        er = "7.5v"
                    elseif ex == 5.0 then
                        if gain_low then
                            ref.max = 50
                            ref.min = -50
                        else
                            ref.max = 4
                            ref.min = -4
                        end
                        er = "5.0v"
                    elseif ex == 2.5 then
                        if gain_low then
                            ref.max = 100
                            ref.min = -100
                        else
                            ref.max = 10
                            ref.min = -10
                        end
                        er = "2.5v"
                    elseif ex == 1.0 then
                        ref.max = 20
                        ref.min = -20
                        er = "1.0v"
                    end

                    s._category_specific[ mx411].excitation_range = er
                    s._category_specific[ mx411].sensor_input.max = ref.max
                    s._category_specific[ mx411].sensor_input.min = ref.min
                    s._category_specific[ mx411].sensor_input.wires = s.sensor_input.wires
                    s._category_specific[ mx411].sensor_input.regulation= s.sensor_input.regulation
                    return true
                else
                    if l.Excitation >= 5.0 then
                        sdb_ex = 5.0
                    elseif l.Excitation >= 2.5 then
                        sdb_ex = 2.5
                    elseif l.Excitation >= 1.0 then
                        sdb_ex = 1.0
                    else
                        return false
                    end

                    if max > 250 then
                        return false
                    elseif max > 100 then
                        gain_low = true
                        ex = 1.0
                    elseif max > 20 then
                        gain_low = true
                        if sdb_ex < 2.5 then
                            ex = sdb_ex
                        else
                            ex = 2.5
                        end
                    elseif max > 8 then
                        ex = 1.0
                    elseif max > 4 then
                        if sdb_ex < 2.5 then
                            ex = sdb_ex
                        else
                            ex = 2.5
                        end
                    else
                        ex = sdb_ex
                    end

                    if s.bridge_resistance < 110 then
                        if ex > 1.0 then
                            ex = 1.0
                        end
                    elseif s.bridge_resistance < 300 then
                        if ex > 2.5 then
                            ex = 2.5
                        end
                    end

                    if ex == 5.0 then
                        ref.max = 4
                        ref.min = -4
                        er = "5.0v"
                    elseif ex == 2.5 then
                        if gain_low then
                            ref.max = 100
                            ref.min = -100
                        else
                            ref.max = 8
                            ref.min = -8
                        end
                        er = "2.5v"
                    elseif ex == 1.0 then
                        if gain_low then
                            ref.max = 250
                            ref.min = -250
                        else
                            ref.max = 20
                            ref.min = -20
                        end
                        er = "1.0v"
                    end

                    s._category_specific[ mx411].excitation_range = er
                    s._category_specific[ mx411].sensor_input.max = ref.max
                    s._category_specific[ mx411].sensor_input.min = ref.min
                    s._category_specific[ mx411].sensor_input.wires = s.sensor_input.wires
                    s._category_specific[ mx411].sensor_input.regulation= s.sensor_input.regulation
                    return true
                end
            end
            ,excitation_mode = function( s ,args)
                local l = args.l
                local em
                if l.ExcitationFrequency == -1 then
                    em = 'dc'
                else
                    em = 'ac_4800_hz_sine'
                end
                s._category_specific[ mx411].excitation_mode = em
                return true
            end
        }
        ,[brg] = {
            sensor_input = function (s, args)
                if s.sensor_input["type"] == "full" or s.sensor_input["type"] == "half" then
                    if s.bridge_resistance >= 80 and s.bridge_resistance < 100 then
                        return false
                    end
                end
                if args.l.SADPara < 25 or args.l.SADPara > 29 or arg.l.SADPara == 27 then
                    if s._category_specific[brg].sensor_input.type == "full" then
                        s._category_specific[brg].sensor_input.wires =  4
                        s._category_specific[brg].sensor_input.regulation =  false
                    elseif s._category_specific[brg].sensor_input.type == "half" then
                        s._category_specific[brg].sensor_input.wires =  3
                        s._category_specific[brg].sensor_input.regulation =  false
                    elseif s._category_specific[brg].sensor_input.type == "quarter" then
                        s._category_specific[brg].sensor_input.wires =  3
                        s._category_specific[brg].sensor_input.regulation =  false
                    end
                else
                    if s._category_specific[brg].sensor_input.type == "full" then
                        if s._category_specific[brg].sensor_input.wires == 6 then
                            s._category_specific[brg].sensor_input.wires =  4
                            s._category_specific[brg].sensor_input.regulation =  false
                        end
                        if s._category_specific[brg].sensor_input.wires ~= 4 then
                            return false
                        end
                    elseif s._category_specific[brg].sensor_input.type == "half" then
                        if s._category_specific[brg].sensor_input.wires == 5 then
                            s._category_specific[brg].sensor_input.wires =  3
                            s._category_specific[brg].sensor_input.regulation =  false
                        end
                        if s._category_specific[brg].sensor_input.wires ~= 3 then
                            return false
                        end
                    elseif s._category_specific[brg].sensor_input.type == "quarter" then
                        if s._category_specific[brg].sensor_input.wires ~= 3 then
                            return false
                        end
                    end
                end
                if s._category_specific[brg].sensor_input.type == "quarter" and 
                        s._category_specific[brg].sensor_input.wires == 3  then
                    s._category_specific[brg].sensor_input.regulation = false
                end
                print("sensor_input brg")
                return true
            end
            ,excitation_range = function (s, args)
                local l = args.l
                local out_of_range = false
                local er = l.Excitation
                local ex
                if er >= 10 then ex = "10.0v"
                elseif er >=5 then ex = "5.0v"
                else ex = er .. "v" 
                end
                if s.bridge_resistance < 240 and ex == "10.0v" then ex = "5.0v" end
                s._category_specific[brg].excitation_range = ex 
                print("range brg")
                return true
            end
            ,excitation_mode = function( s ,args)
                s._category_specific[brg].excitation_mode = 'dc'
                print("mode brg")
                return true
            end
        }
        ,[hls] = {
            sensor_input = function (s, args)
                if s.sensor_input["type"] == "full" or s.sensor_input["type"] == "half" then
                    if s.bridge_resistance >= 80 and s.bridge_resistance < 100 then
                        return false
                    end
                end
                if args.l.SADPara < 25 or args.l.SADPara > 29 or arg.l.SADPara == 27 then
                    if s._category_specific[hls].sensor_input.type == "full" then
                        s._category_specific[hls].sensor_input.wires =  4
                        s._category_specific[hls].sensor_input.regulation =  false
                    elseif s._category_specific[hls].sensor_input.type == "half" then
                        s._category_specific[hls].sensor_input.wires =  3
                        s._category_specific[hls].sensor_input.regulation =  false
                    elseif s._category_specific[hls].sensor_input.type == "quarter" then
                        s._category_specific[hls].sensor_input.wires =  3
                        s._category_specific[hls].sensor_input.regulation =  false
                    end
                else
                    if s._category_specific[hls].sensor_input.type == "full" then
                        if s._category_specific[hls].sensor_input.wires == 6 then
                            s._category_specific[hls].sensor_input.wires =  4
                            s._category_specific[hls].sensor_input.regulation =  false
                        end
                        if s._category_specific[hls].sensor_input.wires ~= 4 then
                            return false
                        end
                    elseif s._category_specific[hls].sensor_input.type == "half" then
                        if s._category_specific[hls].sensor_input.wires == 5 then
                            s._category_specific[hls].sensor_input.wires =  3
                            s._category_specific[hls].sensor_input.regulation =  false
                        end
                        if s._category_specific[hls].sensor_input.wires ~= 3 then
                            return false
                        end
                    elseif s._category_specific[hls].sensor_input.type == "quarter" then
                        if s._category_specific[hls].sensor_input.wires ~= 3 then
                            return false
                        end
                    end
                end
                if s._category_specific[hls].sensor_input.type == "quarter" and 
                        s._category_specific[hls].sensor_input.wires == 3  then
                    s._category_specific[hls].sensor_input.regulation = false
                end
                print("sensor_input hls")
                return true
            end
            ,excitation_range = function (s, args)
                local l = args.l
                local out_of_range = false
                local er = l.Excitation
                local ex
                if er >= 10 then ex = "10.0v"
                elseif er >=5 then ex = "5.0v"
                else ex = er .. "v" 
                end
                if s.bridge_resistance < 240 and ex == "10.0v" then ex = "5.0v" end
                s._category_specific[hls].excitation_range = ex 
                print("range hls")
                return true
            end
            ,excitation_mode = function( s ,args)
                s._category_specific[hls].excitation_mode = 'dc'
                print("mode hls")
                return true
            end
        }
    }
    , get_wires = function (s,l)
        local SensorType_enum, SADPara_enum
        SensorType_enum = l.SensorType
        SADPara_enum = l.SADPara
        if SensorType_enum == 368 or SensorType_enum == 369 then return 4 end    
        if SensorType_enum == 371 or SensorType_enum == 372 then return 3 end    
        if SADPara_enum == 25 then return 3 end
        if SADPara_enum == 26 then return 4 end
        if SADPara_enum == 28 then return 5 end
        if SADPara_enum == 29 then return 6 end
            --this is different for mx devices compared to edaq layers.  
            --Defaulting to mx parameters and change is device specific code.
        if bridge_map[SensorType_enum].type == "full" then
            return 6 
        elseif bridge_map[SensorType_enum].type  == "half" then
            return 5
        elseif bridge_map[SensorType_enum].type == "quarter" then
            return 3
        end
    end
    ,req = function ( s ,l )
        local supported = false
        for k,v in ipairs( supported_bridges ) do
            if v == l.SensorType then
                supported = true
            end
        end
        if supported then
            if l.SensorType == 368 or l.SensorType == 369 then
                s._supported_categories = {mx1615}
            end
        end    
        return supported
    end
    ,setup = function( s , l ) --this is really messy, but the rules are kind of all over the place.
        local t = l.SensorType
        local failed = false
        s._supported_categories = { mx1615 ,mx840, mx411, brg, hls}
        if ss.StrainGage.req( s , l ) then
            s.sensor_input = bridge_map[l.SensorType]
            s.bridge_resistance = bridge_resist_map[l.SensorType]
            s.sensor_input.wires = ss.StrainGage.get_wires(s,l)
            if not s.bridge_resistance then
                s.bridge_resistance = l.Resistance
                if s.bridge_resistance == 0 then s.bridge_resistance = 350 end
            end
            s.bridge_lead_wire_resistance = 0
            
            if s.sensor_input["type"] == "full" or s.sensor_input["type"] == "half" then
                if s.bridge_resistance < 80 then
                    
                    return true
                elseif s.bridge_resistance > 10000 then
                    return true
                elseif s.bridge_resistance > 5000 then
                    s._supported_categories = {brg, hls}
                end
            elseif s.sensor_input["type"] == "quarter" then
                if s.bridge_resistance ~= 120 and s.bridge_resistance ~= 350 then
                    return true
                end
            else --not sure how this would be possible
                return true
            end

            if s.sensor_input["type"] == "full" then
                if s.sensor_input.wires == 6 then s.sensor_input.regluation = true end
                if s.sensor_input.wires == 4 then s.sensor_input.regluation = false end
            elseif s.sensor_input["type"] == "half" then
                if s.sensor_input.wires == 5 then s.sensor_input.regulation = true end
                if s.sensor_input.wires == 3 then s.sensor_input.regulation = false end
            elseif s.sensor_input["type"] == "quarter" then
                if s.sensor_input.wires == 4 then s.sensor_input.regulation = true end
                if s.sensor_input.wires == 3 then s.sensor_input.regulation = true end
            else
                return true
            end

            if setup_category_specifics( 'StrainGage' ,s ,{l=l}) then
                failed = true
                print( 'Failed category specific check: ' ,l.Description)
            end
        else
            failed = true
            print( "Failed general check " .. l.Description)
        end
        --    s.sensor_input = nil
        return failed
    end
}

--Bridge and StrainGage are identical, save for their enums
--I'm keeping them seperate just in case something comes up in the future,
--otherwise I'll eventually merge them and adjust the system accordingly
ss.Bridge = copy_table( ss.StrainGage )
ss.Bridge.name = "Bridge"
ss.Bridge.enum = 2

ss.Voltage = {
    name="Voltage"
    ,enum = 4
    ,elec_units = "V"
    ,s_types = {
        scale_types.ZeroSpan
        ,scale_types.TwoPoint
        ,scale_types.Table2
        ,scale_types.Gain
    }
    ,overlaps = {
        [mx1601] = {
            sensor_input = function( s ,args)
                local max = tonumber(args.max)
                local ref = {}
                if max <= 0.1 then
                    max = 0.1
                elseif max <= 10 then
                    max = 10
                elseif max <= 60 then
                    max = 60
                else
                    max = nil
                end
                if max then
                    ref.max = max
                    ref.min = max * -1
                    s._category_specific[ mx1601].sensor_input.max = ref.max
                    s._category_specific[ mx1601].sensor_input.min = ref.min
                    return true
                else
                    return false
                end
            end,
            transducer_power = function(s, args)
                s._category_specific[mx1601].transducer_power = s.transducer_power 
                return true 
            end
        }
        ,[mx840] = {
            sensor_input = function( s ,args)
                local max = tonumber(args.max)
                local ref = {}
                if max <= 0.1 then
                    max = 0.1
                elseif max <= 10 then
                    max = 10
                elseif max <= 60 then
                    max = 60
                elseif max <= 300 then
                    max = 300

                else
                    max = nil
                end
                if max then
                    ref.max = max
                    ref.min = max * -1
                    s._category_specific[ mx840].sensor_input.max = ref.max
                    s._category_specific[ mx840].sensor_input.min = ref.min
                    return true
                else
                    return false
                end
            end,
            transducer_power = function(s, args)
                s._category_specific[mx840].transducer_power = s.transducer_power 
                return true
            end
        }
        ,[mx1615] = {
            sensor_input = function( s ,args)
                local max = tonumber(args.max)
                local ref = {}
                if max >= 0.6 and max <= 60 then
                    max = 15
                else
                    max = nil
                end
                if max then
                    ref.max = max
                    ref.min = max * -1
                    s._category_specific[ mx1615].sensor_input.max = ref.max
                    s._category_specific[ mx1615].sensor_input.min = ref.min
                    return true
                else
                    return false
                end
            end,
            transducer_power = function(s, args)
                s._category_specific[mx1615].transducer_power = s.transducer_power 
                return true
            end
        }
        ,[mx411] = {
            sensor_input = function( s ,args)
                local max = tonumber(args.max)
                local ref = {}
                if max >= 0.1 and max <= 10 then
                    ref.max = 10
                    ref.min = -10
                    s._category_specific[ mx411].sensor_input.max = ref.max
                    s._category_specific[ mx411].sensor_input.min = ref.min
                    return true
                else
                    return false
                end
            end,
            transducer_power = function(s, args)
                s._category_specific[mx411].transducer_power = s.transducer_power 
                return true
            end
        }
        ,[hls] = {
            sensor_input = function( s ,args)
                local max = tonumber(args.max)
                local ref = {}
                if max < .1 then 
                    return false
                elseif max <= 10 then
                    ref.max = 10
                    ref.min = -10
                    s._category_specific[ hls].sensor_input.max = ref.max
                    s._category_specific[ hls].sensor_input.min = ref.min
                    return true
                elseif max <= 80 then
                    ref.max = 80
                    ref.min = -80
                    s._category_specific[ hls].sensor_input.max = ref.max
                    s._category_specific[ hls].sensor_input.min = ref.min
                    return true
                else
                    print("max not supported:\t" .. max)
                    return false
                end
            end,
            transducer_power = function(s, args)
                s._category_specific[hls].transducer_power = math.floor(s.transducer_power) 
                return true
            end
        }
        ,[brg] = {
            sensor_input = function( s ,args)
                local max = tonumber(args.max)
                local ref = {}
                if max < .1 then 
                    return false
                elseif max <= 10 then
                    ref.mode = "volt"
                    ref.max = 10
                    ref.min = -10
                    s._category_specific[ brg].sensor_input.max = ref.max
                    s._category_specific[ brg].sensor_input.min = ref.min
                    return true
                else
                    return false
                end
            end,
            transducer_power = function(s, args)
                s._category_specific[brg].transducer_power = s.transducer_power 
                return true
            end
        }
    }
    ,req = function( s , l )
        if not tonumber(l.ActiveSupplyVoltage) then
            l.ActiveSupplyVoltage = 0
        else
            l.ActiveSupplyVoltage = tonumber(l.ActiveSupplyVoltage)
        end
        if l.SensorType == 420 then
            if l.ActiveSupplyVoltage == 0 then
                return true
            elseif l.ActiveSupplyVoltage >= 3 then
                if l.ActiveSupplyVoltage >= 5 then
                    if l.ActiveSupplyVoltage <= 24 then
                        s._supported_categories = {mx411, mx840, mx1601, hls}
                        --cut out 1615,brg since invalid in this range
                        return true
                    elseif l.ActiveSupplyVoltage < 29 then
                        s._supported_categories = { mx1601, hls}
                        return true
                    elseif l.ActiveSupplyVoltage <= 30 then
                        s._supported_categories = { mx1601}
                        --cut out 1615, 411, and 840,  since invalid in this range
                        return true
                    else 
                        return false
                    end
                else
                    s._supported_categories = {hls}
                    --no 1601,1615,840,411,brg
                    return true
                end
            end
        end
        return false
    end
    ,setup = function( s , l )
        local failed = false
        s._supported_categories  = { mx1601 , mx840, mx411, mx1615, brg, hls }

        if ss.Voltage.req( s , l ) then
            s.transducer_power = l.ActiveSupplyVoltage
            s.sensor_input = {}
            s.sensor_input.mode = "volt"
            s.calibration[3] = nil
            print("supported_categories")
            for k,v in ipairs(s._supported_categories) do
                print(k .. ":\t" .. v)
            end

            local max = 0
            if l.ScaleType == 1 then --ZeroSpan
                print("span:\t" .. s.calibration[1].electrical_span .. "\telectrical zero:\t".. s.calibration[1].electrical_zero) 
                max = math.abs(s.calibration[1].electrical_span) + math.abs(s.calibration[1].electrical_zero)
                print("max:\t" .. max)
            elseif l.ScaleType == 2 or l.ScaleType == 32 then --Table2/TwoPoint
                max = s.calibration[1].electrical[2]
            elseif l.ScaleType == 64 then --Gain
                max = l.Range
            end

            if setup_category_specifics( 'Voltage' ,s ,{max=max}) then
                failed = true
                print( 'Failed category specific: ' ,l.Description)
            end

        else
            print( "Failed general check " .. l.Description )
            failed = true
        end
        return failed
    end
}

ss.Current = {
    name="Current"
    ,enum = 5
    ,elec_units = "mA"
    ,s_types = {
        scale_types.ZeroSpan
        ,scale_types.TwoPoint
        ,scale_types.Table2
        ,scale_types.Gain
    }
    ,req = function( s , l )
        if not tonumber(l.ActiveSupplyVoltage) then
            l.ActiveSupplyVoltage = 0
        else
            l.ActiveSupplyVoltage = tonumber(l.ActiveSupplyVoltage)
        end
        return l.SensorType == 421
            and ( (l.ActiveSupplyVoltage >= 5
            and l.ActiveSupplyVoltage <= 30)
            or l.ActiveSupplyVoltage == 0 )

    end
    ,setup = function( s , l )
        local failed = false

        s._supported_categories = { mx1601, mx840, mx411 }
        if ss.Current.req( s , l ) then
            s.sensor_input.mode = "4-20ma"
            s.transducer_power = l.ActiveSupplyVoltage
            s.calibration[3] = nil
        else
            failed = true
            print( "Failed general check " .. l.Description )
        end
        return failed
    end
}

ss.Thermocouple = {
    name="Thermocouple"
    ,enum = 8
    ,elec_units = "°C"
    ,s_types = {
        scale_types.Internal
    }
    ,overlaps = {
        [mx1609] = {
            sensor_input = function (s, args)
                local l = args.l
                if l.SensorType ~= 451 then
                    return false
                else
                    s._category_specific[mx1609].sensor_input = {}
                    s._category_specific[mx1609].sensor_input.type = "k"
                    s._category_specific[mx1609].sensor_input.mode = "thermocouple"
                    return true
                end
            end
        }
        ,[mx840] = {
            sensor_input = function(s, args)
                local l = args.l
                local thermType = l.SensorType
                s._category_specific[mx840].sensor_input = {}
                s._category_specific[mx840].sensor_input.teds_adapter_required = true
                s._category_specific[mx840].sensor_input.mode = "thermocouple"
                if thermType == 450 then
                    s._category_specific[mx840].sensor_input.type = "j"
                elseif thermType == 451 then 
                    s._category_specific[mx840].sensor_input.type = "k"
                elseif thermType == 452 then 
                    s._category_specific[mx840].sensor_input.type = "t"
                elseif thermType == 453 then 
                    s._category_specific[mx840].sensor_input.type = "s"
                elseif thermType == 454 then 
                    s._category_specific[mx840].sensor_input.type = "b"
                elseif thermType == 455 then 
                    s._category_specific[mx840].sensor_input.type = "e"
                elseif thermType == 456 then 
                    s._category_specific[mx840].sensor_input.type = "r"
                elseif thermType == 457 then 
                    s._category_specific[mx840].sensor_input.type = "n"
                end
                return true
            end
        }
        ,[itb] = {
            sensor_input = function(s, args)
                local l = args.l
                local thermType = l.SensorType
                if thermType > 452 and thermType ~= 455 then
                    return false
                end
                s._category_specific[itb].sensor_input = {}
                s._category_specific[itb].sensor_input.do_invalid_data_fills = true
                s._category_specific[itb].sensor_input.mode = "thermocouple"
                if thermType == 450 then
                    s._category_specific[itb].sensor_input.type = "j"
                elseif thermType == 451 then 
                    s._category_specific[itb].sensor_input.type = "k"
                elseif thermType == 452 then 
                    s._category_specific[itb].sensor_input.type = "t"
                elseif thermType == 455 then 
                    s._category_specific[itb].sensor_input.type = "e"
                end
                return true 
            end
        }
        ,[ntb] = {
            sensor_input = function(s, args)
                local l = args.l
                local thermType = l.SensorType
                if thermType > 452 and thermType ~= 455 then
                    return false
                end
                s._category_specific[ntb].sensor_input = {}
                s._category_specific[ntb].sensor_input.do_invalid_data_fills = true
                s._category_specific[ntb].sensor_input.mode = "thermocouple"
                if thermType == 450 then
                    s._category_specific[ntb].sensor_input.type = "j"
                elseif thermType == 451 then 
                    s._category_specific[ntb].sensor_input.type = "k"
                elseif thermType == 452 then 
                    s._category_specific[ntb].sensor_input.type = "t"
                elseif thermType == 455 then 
                    s._category_specific[ntb].sensor_input.type = "e"
                end
                return true 
            end
        }
        ,[hls] = {
            sensor_input = function(s, args)
                local l = args.l
                local thermType = l.SensorType
                if thermType > 452 and thermType ~= 455 then
                    return false
                end
                s._category_specific[hls].sensor_input = {}
                s._category_specific[hls].sensor_input.mode = "thermocouple"
                if thermType == 450 then
                    s._category_specific[hls].sensor_input.type = "j"
                elseif thermType == 451 then 
                    s._category_specific[hls].sensor_input.type = "k"
                elseif thermType == 452 then 
                    s._category_specific[hls].sensor_input.type = "t"
                elseif thermType == 455 then 
                    s._category_specific[hls].sensor_input.type = "e"
                end
                return true 
            end
        }
    }
    ,req = function( s , l )
        return l.SensorType >= 450 and l.SensorType <= 457 and
            ( l.EngUnit:match( '°C') or l.EngUnit:match('°F') or l.EngUnit:match('°R'))
    end
    ,setup = function( s , l )
        local failed = false

        s._supported_categories = { mx1609, mx840, ntb, itb, hls}
        if ss.Thermocouple.req( s , l ) then
            --print( s.source_physical_units
            --    ,s.electrical_units , s.output_units )
            s.sensor_input.mode = "thermocouple"
            s.sensor_input.type = 'k'
            s.source_physical_units = '°C'
        else
            failed = true
            print( "Thermo Failed general check "
                .. l.Description , l.EngUnit , l.SensorType )
        end
        if setup_category_specifics('Thermocouple' , s , {l=l}) then
            failed = true
        end
        return failed
    end

}

ss.RTD = {
    name="RTD"
    ,enum = 7
    ,elec_units = "°C"
    ,s_types = {
        scale_types.Internal
    }
    ,req = function( s , l )
        return (l.SensorType == 503 or l.SensorType == 501) and l.SADPara == 26 and
            ( l.EngUnit:match( '°C') or l.EngUnit:match('°F') or l.EngUnit:match('°R'))
    end
    ,setup = function( s , l )
        local failed = false
        s._supported_categories = { mx840, mx1615 }

        if ss.RTD.req( s , l ) then
            if l.SensorType == 501 then
                s.sensor_input.mode = "pt100"
            elseif l.SensorType == 503 then
                table.remove(s._supported_categories, 2)
                --cut out 1615, since invalid for the pt1000
                s.sensor_input.mode = "pt1000"
            end
        else
            failed = true
            print( "RTD Failed general check "
                .. l.Description , l.EngUnit , l.SensorType )
        end
        return failed
    end

}

ss.Resistance =     {
    name="Resistance"
    ,enum = 10
    ,elec_units = "Ohm"
    ,s_types = {
        scale_types.ZeroSpan
        ,scale_types.TwoPoint
        ,scale_types.Table2
        ,scale_types.Gain
    }
    ,req = function( s , l )
        return l.SensorType == 475 and l.SADPara == 26
    end
    ,setup = function( s , l )
        local failed = false
        s._supported_categories = { mx1615, mx840}

        if ss.Resistance.req( s , l ) then
            s.sensor_input.mode = "resistor"
            s.calibration[3] = nil
        else
            failed = true
            print( "Failed general check")
        end
        return failed
    end
}

ss.Potentiometric =     {
    name="Potentiometric"
    ,enum = 14
    ,elec_units = "mV/V"
    ,s_types = {
        scale_types.ZeroSpan
        ,scale_types.TwoPoint
        ,scale_types.Table2
        ,scale_types.Gain
    }
    ,overlaps = {
        [mx1615] = {
            excitation_range = function( s ,args)
                s._category_specific[ mx1615].excitation_range = "1.0v"
                return true
            end
        }
        ,[mx840] = {
            excitation_range = function( s ,args)
                s._category_specific[ mx840].excitation_range = "2.5v"
                return true
            end
        }
    }

    ,req = function( s , l )
        return l.SensorType == 385 and l.SADPara == 28
    end

    ,setup = function( s , l )
        local failed = false
        s._supported_categories = { mx1615, mx840}

        if ss.Potentiometric.req( s , l ) then
            s._supported_categories = ss.Potentiometric.categories
            s.sensor_input.mode = 'potentiometer'
            s.calibration[3] = nil
            s.excitation_mode = "dc"

        else
            failed = true
            print( "Failed general check" ,l.SensorType , l.SADPara)
        end
        if setup_category_specifics( "Potentiometric" ,s ,{l=l}) then
            failed = true
            print( 'Failed category specific: ' ,l.Description)
        end
        return failed
    end

}

ss.IEPE = {
    name="IEPE"
    ,enum = 31
    ,elec_units = "V"
    ,s_types = {
        scale_types.ZeroSpan
        ,scale_types.TwoPoint
        ,scale_types.Table2
        ,scale_types.Gain
    }
    ,req = function( s , l )
        return l.SensorType == 581
    end
    ,setup = function( s , l )
        local failed = false
        s._supported_categories = { mx1601, mx840, mx411, hls }

        if ss.IEPE.req( s , l ) then
            s._supported_categories = ss.IEPE.categories
            s.sensor_input.mode = "iepe"
            s.calibration[3] = nil
        else
            failed = true
        end
        return failed
    end
}

ss.IndBridge = {
    name="IndBridge"
    ,enum = 11
    ,elec_units = "mV/V"
    ,s_types = {
        scale_types.ZeroSpan
        ,scale_types.TwoPoint
        ,scale_types.Table2
        ,scale_types.Gain
    }
    ,overlaps = {
        [mx840] = {
            excitation_range = function( s ,args)
                local l = args.l
                local max
                local ex, sdb_ex, gain_low, gain_low_low
                local ref = {}
                if l.ScaleType == 0 or l.ScaleType == 64 then --StrainGage or Gain
                    max = l.Range
                elseif l.ScaleType == 1 then --ZeroSpan
                    max = math.abs(l.P2X-l.P1X) + math.abs(l.P1X)
                elseif l.ScaleType == 2 or l.ScaleType == 32 then -- Table2 or  TwoPoint
                    max = l.P2X
                else
                    return false
                end
                sdb_ex = l.Excitation
                if sdb_ex >= 2.5 then
                    sdb_ex = 2.5
                elseif sdb_ex >= 1.0 then
                    sdb_ex = 1.0
                else
                    return false
                end
                if l.SensorType == 357 and max > 300 then --half bridge
                    return false
                end
                gain_low = false
                gain_low_low = false
                if max > 1000 then
                --    print("Failed excitation range " .. l.Description )
                    return false
                elseif max > 300 then
                    gain_low_low = true
                    ex = "1.0v"
                elseif max > 120 then
                    ex = "1.0v"
                    gain_low = true
                elseif max > 12.5 then  
                    gain_low = true
                    if sdb_ex < 2.5 then
                        ex = tostring(sdb_ex) .. "v"
                    else
                        ex = "2.5v"
                    end
                elseif max > 5 then
                    ex = "1.0v" 
                else
                    ex = tostring(sdb_ex) .. "v"
                end

                if ex == "2.5v" then
                    if gain_low then
                        ref.max = 120
                    else
                        ref.max = 5
                    end
                else
                    if gain_low_low then
                        ref.max = 1000
                    elseif gain_low then
                        ref.max = 300
                    else 
                        ref.max = 12.5
                    end
                end

                ref.min = -1 * ref.max
                ref.mode = s.sensor_input.mode
                ref.type = s.sensor_input.type
                ref.wires = s.sensor_input.wires
                s._category_specific[ mx840].sensor_input = ref
                s._category_specific[ mx840].excitation_range = ex
                return true
            end
        }
        ,[mx411] = {
            excitation_range = function( s ,args)
                local l = args.l
                local max
                local ex, sdb_ex, gain_low, gain_low_low
                local ref = {}
                if l.ScaleType == 0 or l.ScaleType == 64 then --StrainGage or Gain
                    max = l.Range
                elseif l.ScaleType == 1 then --ZeroSpan
                    max = math.abs(l.P2X-l.P1X) + math.abs(l.P1X)
                elseif l.ScaleType == 2 or l.ScaleType == 32 then -- Table2 or  TwoPoint
                    max = l.P2X
                else
                    return false
                end
                sdb_ex = l.Excitation
                if sdb_ex >= 2.5 then
                    sdb_ex = 2.5
                elseif sdb_ex >= 1.0 then
                    sdb_ex = 1.0
                else
                    return false
                end
                
                gain_low = false
                gain_low_low = false
                if max > 250 then
                    return false
                elseif max > 100 then
                    gain_low_low = true
                     ex = "1.0v"
                elseif max > 20 then  
                    gain_low = true
                    if sdb_ex < 2.5 then
                        ex = tostring(sdb_ex) .. "v"
                    else
                        ex = "2.5v"
                    end
                elseif max > 8 then
                    ex = "1.0v" 
                else
                    ex = tostring(sdb_ex) .. "v"
                end

                if ex == "2.5v" then
                    if gain_low then
                        ref.max = 100
                    else
                        ref.max = 8
                    end
                else
                    if gain_low then
                        ref.max = 250
                    else 
                        ref.max = 20
                    end
                end

                ref.min = -1 * ref.max
                ref.mode = s.sensor_input.mode
                ref.type = s.sensor_input.type
                ref.wires = s.sensor_input.wires
                s._category_specific[ mx411].sensor_input = ref
                s._category_specific[ mx411].excitation_range = ex
                return true
            end
        }
    }
    ,req = function( s , l )
        return l.SensorType == 356 or l.SensorType == 357
    end
    ,setup = function( s , l )
        local failed = false
        s._supported_categories = { mx840, mx411 }

        if ss.IndBridge.req( s , l ) then
            s.calibration[3] = nil
            s.excitation_mode = "ac_4800_hz_sine"
            if l.SensorType == 356 then
                s.sensor_input.mode = "inductive_bridge"
                s.sensor_input.type = "full"
                s.sensor_input.wires = "6"
            else
                s.sensor_input.mode = "inductive_bridge"
                s.sensor_input.type = "half"
                s.sensor_input.wires = "5"

            end
            if setup_category_specifics( 'IndBridge' ,s ,{l=l}) then
                failed = true
                print( 'Failed category specific check: ' ,l.Description)
            end
        else
            failed = true
        end
        return failed
    end
}

ss.LVDT = {
    name="LVDT"
    ,enum = 3
    ,elec_units = "mV/V"
    ,s_types = {
        scale_types.ZeroSpan
        ,scale_types.TwoPoint
        ,scale_types.Table2
        ,scale_types.Gain
    }
    ,req = function( s , l )
        return l.SensorType == 380
    end
    ,setup = function( s , l )
        local failed = false

        s._supported_categories = { mx840}
        if ss.LVDT.req( s , l ) then
            s._supported_categories = ss.LVDT.categories
            s.sensor_input.mode = "lvdt"
            s.excitation_range = "1.0v"
            s.excitation_mode = "ac_4800_hz_sine"
            s.calibration[3] = nil
        else
            failed = true
        end
        return failed
    end
}

ss.Frequency = {
    name="Frequency"
    ,enum = 6
    ,elec_units = "kHz"
    ,s_types = {
        scale_types.ZeroSpan
        ,scale_types.TwoPoint
        ,scale_types.Table2
        ,scale_types.Gain
    }
    ,req = function( s , l )
        local voltage
        if not tonumber(l.ActiveSupplyVoltage) then
            voltage = 0
            l.ActiveSupplyVoltage = 0
        else
            voltage = tonumber(l.ActiveSupplyVoltage)
        end
        if voltage == 0 or (voltage >= 5 and voltage <= 24) then
            s.transducer_power = voltage
            return l.SensorType == 520 or l.SensorType == 522 or l.SensorType == 523
        else
            return false
        end
    end
    ,setup = function( s , l )
        local failed = false
        s._supported_categories = { mx840 }
        if ss.Frequency.req( s , l ) then
            s.calibration[3] = nil
            if l.SensorType == 520 then
                s.sensor_input.mode = "pulse_frequency"
            elseif l.SensorType == 522 then
                s.sensor_input.mode = "encoder_frequency"
                s.sensor_input.quadrature = "false"
            else
                s.sensor_input.mode = "encoder_frequency"
                s.sensor_input.quadrature = "true"
            end
        else
            failed = true
        end
        return failed
    end
}

ss.Incremental = {
    name="Incremental"
    ,enum = 12
    ,elec_units = "kImp"
    ,s_types = {
        scale_types.ZeroSpan
        ,scale_types.TwoPoint
        ,scale_types.Table2
        ,scale_types.Gain
    }
    ,req = function( s , l )
        if not tonumber(l.ActiveSupplyVoltage) then
            l.ActiveSupplyVoltage = 0
        else
            l.ActiveSupplyVoltage = tonumber(l.ActiveSupplyVoltage)
        end
        if l.ActiveSupplyVoltage == 0 or (l.ActiveSupplyVoltage >= 5 and l.ActiveSupplyVoltage <= 24) then
            return l.SensorType == 525 or l.SensorType == 522 or l.SensorType == 523
        else
            print("Sensor type not supported")
            return false
        end
    end
    ,setup = function( s , l )
        local failed = false
        s._supported_categories = { mx840 }
        --print("Counter")
        if ss.Incremental.req( s , l ) then
            s._supported_categories = ss.Incremental.categories
            s.transducer_power = l.ActiveSupplyVoltage
            s.calibration[3] = nil
            if l.SensorType == 525 then
                s.sensor_input.mode = "pulse_counter"
            else
                s.sensor_input.mode = "encoder_counter"
                if l.SensorType == 522 then
                    s.sensor_input.quadrature = "false"
                else
                    s.sensor_input.quadrature = "true"
                end
                if l.CounterIndexEnabled == "null" then
                    l.CounterIndexEnabled = 0
                end
                if not tonumber(l.CounterIndexDivisor) then
                    l.CounterIndexDivisor = 1
                end
                if l.CounterIndexEnabled and (l.CounterIndexDivisor >= 1 or l.CounterIndexDivisor <= 4) then
                    if l.CounterIndexEnabled == 1 then
                        s.sensor_input.encoder_index_reset = true
                        s.sensor_input.encoder_index_divisor = l.CounterIndexDivisor
                    elseif l.CounterIndexEnabled == 0 then
                        s.sensor_input.encoder_index_reset = false
                    else
                        print("sensor index failed")
                        failed = true
                    end
                else
                    print("counter index failed")
                    failed = true
                end
            end
        else
            print("req failed ")
            failed = true
        end
        return failed
    end
}
ss.SSI =     {
    name="SSI"
    ,enum = 30
    ,elec_units = "Imp"
    ,s_types = {
        scale_types.ZeroSpan
        ,scale_types.TwoPoint
        ,scale_types.Table2
        ,scale_types.Gain
    }
    ,req = function( s , l )
        local voltage
        if not tonumber(l.ActiveSupplyVoltage) then
            l.ActiveSupplyVoltage = 0
        end
        voltage = tonumber(l.ActiveSupplyVoltage)
        if voltage == 0 or (voltage >= 5 and voltage <= 24) then
            s.transducer_power = voltage
            return l.SensorType == 580
        else
            return false
        end
    end
    ,setup = function( s , l )
        local failed = false
        s._supported_categories = {mx840}
        --print("req: " .. tostring(ss.SSI.req( s , l )))
        if ss.SSI.req( s , l ) then
            s._supported_categories = ss.SSI.categories
            s.sensor_input.mode = "ssi_encoder_counter"
            --print("ssi coding: " .. tostring(l.SSICoding))
            --print("clock: " .. tostring(l.SSIClockFrequency))
            --print("bits: " .. tostring(l.SSIResolution))
            if l.SSICoding == 0 then
                s.sensor_input.ssi_coding = "gray"
            else
                s.sensor_input.ssi_coding = "binary"
            end
            if l.SSIClockFrequency == 0 then
                s.sensor_input.ssi_clock = 100
            elseif l.SSIClockFrequency == 1 then
                s.sensor_input.ssi_clock = 200
            elseif l.SSIClockFrequency == 2 then
                s.sensor_input.ssi_clock = 500
            elseif l.SSIClockFrequency == 3 then
                s.sensor_input.ssi_clock = 1000
            end
            s.sensor_input.ssi_bits = tonumber(l.SSIResolution) + 12
            s.calibration[3] = nil
        else
            failed = true
            print( "Failed general check")
        end
        return failed
    end
}

return ss


