-- This contains a date and time setting utility function.
-- The function is used by the API server and by the GPS proxy.

local armada = require 'armada'

m = {}

-- Sets both time domains to the given date and time value.
-- Note that this only sets the time with one second precision.
-- Param:  new_date_time - a table containing the new date
--             and time values, as needed by the Lua function
--             os.time
--         adjuster - a string identifying the adjusting code
-- Returns:  true if successful, false otherwise
function m.set_datetime(new_date_time, adjuster)
    assert(type(new_date_time) == 'table' and type(adjuster) == 'string',
        "Internal error:  bad input parameter")

    local dt_str = string.format("%4d.%02d.%02d-%02d:%02d:%02d",
        new_date_time.year, new_date_time.month, new_date_time.day,
        new_date_time.hour, new_date_time.min, new_date_time.sec)
    local offset = os.time(new_date_time) - os.time(os.date('*t'))
    local cmd = string.format("date -u -s %s", dt_str)
    local status = os.execute(cmd)
    if status ~= -1 then
        -- If configuratiion enables setting of the time stamping hardware time
        -- when the date/time are set, then set it too!
        armada.timing:AdjustClock(offset, 0, adjuster)
    end
    return status ~= -1
end

return m
